/* ==================================================== */
/* Gloup Mar. 92 					*/
/* Copyright (c) Atelier de BioInformatique		*/
/* @file: set.c						*/
/* 'set' management					*/
/* ====================================================	*/
  
#include <stdio.h>
#include <string.h>
#include "../../include/gtypes.h"
#include "../../include/libset.h"




/* ------------------------------------------------------------ */
/* hashcode table to computeCard(E) 				*/
/* hash[i] = # bits to 1 in i       				*/
/* (produced by cardh.c)            				*/
/* ------------------------------------------------------------ */

static Int32 CardHTable[] = {


#include "../../include/cardhtable.h"



};

/* ------------------------------------------------------------ */
/* built a new set E and set it to empty set			*/
/* E = NEW = {}							*/
/* ------------------------------------------------------------ */

SetPtr NewSet(dim)
	Int32 dim;
{
	Int32	i;
	ULong   *pe;
	SetPtr	E;

	E = NEW(Set);

	E->size = (dim >> kLog2BitLong) + 1;

	E->dim = dim;

	pe = E->bits = NEWN(unsigned long, E->size);

	for (i = 0 ; i < E->size ; i++)
	    *pe++ = 0L;

	return E;
}

/* ------------------------------------------------------------ */
/* free a previously built E					*/
/* DESTROY E							*/
/* ------------------------------------------------------------ */

void FreeSet(E)
	SetPtr	E;
{
	if (E) {
	   if (E->bits)
		FREE(E->bits);
	   FREE(E);
	}
}

/* ------------------------------------------------------------ */
/* add the element x (index) into E				*/
/* E = E U {x}							*/
/* ------------------------------------------------------------ */		
void AddElt(E, x)
	SetPtr  E;
	Int32	x;
{
	Int32 q, r;
	ULong mask;
	
	q = x >> kLog2BitLong;
	r = x % kBitsPerLong;

	mask = 0x1L << (kMaxShftLong - r);

	E->bits[q] |= mask;
}

/* ------------------------------------------------------------ */
/* remove the element x (index) from E				*/
/* E = E - {x}							*/
/* ------------------------------------------------------------ */
		
void RemElt(E, x)
	SetPtr  E;
	Int32	x;
{
	Int32 q, r;
	ULong  mask;
	
	q = x >> kLog2BitLong;
	r = x % kBitsPerLong;

	mask = 0x1L << (kMaxShftLong - r);

	E->bits[q] &= ~mask;
}

/* ------------------------------------------------------------ */
/* copy a set E into a set F (same size)			*/
/* F = E							*/
/* ------------------------------------------------------------ */

void CopySet(F, E)
	SetPtr F, E;
{
	Int32 i;
	ULong *pf, *pe;

	pf = F->bits;
	pe = E->bits;

	for (i = 0 ; i < F->size ; i++)
	    *pf++ = *pe++;
}

/* ------------------------------------------------------------ */
/* empty a set E						*/
/* E = {}							*/
/* ------------------------------------------------------------ */
		
void EmptySet(E)
	SetPtr  E;
{
	Int32 i;
	ULong *pe;

	pe = E->bits;

	for (i = 0 ; i < E->size ; i++)
	    *pe++ = 0L;
}

/* ------------------------------------------------------------ */
/* compute the union of two sets (same size)			*/
/* U = E Union F						*/
/* ------------------------------------------------------------ */
		
void UnionSet(U, E, F)
	SetPtr  U, E, F;
{
	Int32 i;
	ULong *pu, *pe, *pf;

	pu = U->bits;
	pe = E->bits;
	pf = F->bits;

	for (i = 0 ; i < U->size ; i++)
	    *pu++ = *pe++ | *pf++;
}

/* ------------------------------------------------------------ */
/* compute the intersection of two sets (same size)		*/
/* I = E Union F						*/
/* ------------------------------------------------------------ */
		
void InterSet(I, E, F)
	SetPtr  I, E, F;
{
	Int32 i;
	ULong *pi, *pe, *pf;

	pi = I->bits;
	pe = E->bits;
	pf = F->bits;

	for (i = 0 ; i < I->size ; i++)
	    *pi++ = *pe++ & *pf++;
}

/* ------------------------------------------------------------ */
/* compute the complement of E in the universe			*/
/* E = Comp(E, Omega)						*/
/* ------------------------------------------------------------ */
		
void ComplemSet(E)
	SetPtr  E;
{
	Int32   i, r;
	ULong mask, *pe;

	pe   = E->bits;

	for (i = 1 ; i < E->size ; i++)
	    *pe++ = ~*pe;

	/* last long  (bits) should be partially complemented	*/

	r = E->dim % kBitsPerLong;
	
	mask = kUn << (kBitsPerLong - r);

	*pe = (~*pe) & mask;	
}


/* ------------------------------------------------------------ */
/* check if E is empty						*/
/* ?(E == {})							*/
/* ------------------------------------------------------------ */
		
Bool IsEmptySet(E)
	SetPtr  E;
{
	Int32   i;
	ULong *pe;

	pe = E->bits;

	for (i = 0 ; i < E->size ; i++)
		if (*pe++) return Faux;

	return Vrai;
}

/* ------------------------------------------------------------ */
/* check if set E is included in set F				*/
/* ?(E c F)							*/
/* ------------------------------------------------------------ */
		
Bool IsIncluded(E, F)
	SetPtr  E, F;
{
	Int32 i;
	ULong *pe, *pf;

	pe = E->bits;
	pf = F->bits;

	for (i = 0 ; i < E->size ; i++)
	    if (((~*pe++) | *pf++) != kUn)
		return Faux;

	return Vrai;
}
	
/* ------------------------------------------------------------ */
/* check if element x (index) belongs to E			*/
/* ?({x} c E)							*/
/* ------------------------------------------------------------ */
		
Bool BelongSet(E, x)
	SetPtr  E;
	Int32	x;
{
	Int32 q, r;
	unsigned long mask;
	
	q = x >> kLog2BitLong;
	r = x % kBitsPerLong;

	mask = 0x1L << (kMaxShftLong - r);

	return (((E->bits[q] & mask) != 0L) ? Vrai : Faux);
}

/* ------------------------------------------------------------ */
/* returns cardinal of E					*/
/* Card(E)							*/
/* ------------------------------------------------------------ */
Int32 CardSet(E, x)
	SetPtr  E;
	Int32	x;
{
	Int32 i, bsize, card;
	unsigned char *bbits;

	bbits = (unsigned char *)E->bits;
	bsize = E->size * sizeof(unsigned long);
	
	for (i = card = 0 ; i < bsize ; i++)
	    card += CardHTable[*bbits++];

	return card;
}

/* ------------------------------------------------------------ */
/* print a set (Debug purpose)					*/
/* ------------------------------------------------------------ */

void PrintSet(stream, E)
	FILE   *stream;
	SetPtr E;
{
	Int32 i;
	
	fprintf(stream, "{ ");

	for (i = 0 ; i < E->dim ; i++)
	    if (BelongSet(E, i))
		fprintf(stream, "%d ", i);

	fprintf(stream, "}\n");
} 
