/* ==================================================== */
/*	Copyright (c) Atelier de BioInformatique	*/ 
/*	Mar. 92 					*/
/* 	File: libstkp.c					*/
/*	Purpose: A library to deal with 'stacks' of 	*/
/*		 pointers 				*/
/*	Note: 'stacks' are dynamic (i.e. size is 	*/
/*	      automatically readjusted when needed)	*/
/*	Note: this is basically the same listing as 	*/
/*	      libstki.c but for pointers		*/
/*	History:					*/
/*	00/03/92 : <Gloup> first draft			*/
/*	15/08/93 : <Gloup> revised version		*/
/* ==================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../../include/gtypes.h"
#include "../../include/libstkp.h"


/* ============================ */
/* Constantes et Macros locales	*/
/* ============================	*/

#define ExpandStack(stkh) ResizeStackp((stkh), (*stkh)->size << 1)

#define ShrinkStack(stkh) ResizeStackp((stkh), (*stkh)->size >> 1)


static Int16 sStkpLastError = kStkpNoErr;

/* -------------------------------------------- */
/* gestion des erreurs			 	*/
/* get/reset erreur flag			*/
/*						*/
/* @function: StkpError				*/
/* -------------------------------------------- */

Int16 StkpError(reset)
	Bool reset;
{
	Int16 err;
	
	err = sStkpLastError;
	
	if (reset)
	   sStkpLastError = kStkpNoErr;
	   
	return err;
	
} /* end of StkpError */

/* -------------------------------------------- */
/* creation d'un stack			 	*/
/*						*/
/* @function: NewStackp				*/
/* -------------------------------------------- */

StackpPtr NewStackp(size)
	Int32 size;
{
	StackpPtr stkp;
	
	if (! (stkp = NEW(Stackp)))
	    return NULL;

	stkp->size    = size;
	stkp->top     = 0;
	stkp->cursor  = 0;

	if ( ! (stkp->val = NEWN(Ptr, size))) 
        {
	    sStkpLastError = kStkpMemErr;
	    return FreeStackp(stkp);
	}

	return stkp;	
			
} /* end of NewStackp */


/* -------------------------------------------- */
/* liberation d'un stack			*/
/* use freeFunc to free internal components	*/
/*						*/
/* @function: FreeStackp			*/
/* -------------------------------------------- */

StackpPtr FreeStackp(stkp)
	StackpPtr stkp;
{
	if (stkp) {
	    if (stkp->val)
	    	FREE(stkp->val);
	    FREE(stkp);
	}
	
	return NULL;
	
} /* end of FreeStackp */


/* -------------------------------------------- */
/* creation d'un vecteur de stacks		*/
/*						*/
/* @function: NewStackpVector			*/
/* -------------------------------------------- */

StackpHdle NewStackpVector(vectSize, stackSize)
	Int32 vectSize, stackSize;
{
	Int32	 	i;
	StackpHdle	stkh;
	
	if (! (stkh = NEWN(StackpPtr, vectSize))) {
	    sStkpLastError = kStkpMemErr;
	    return NULL;
	}
	
	for (i = 0 ; i < vectSize ; i++)
	    if (! (stkh[i] = NewStackp(stackSize)))
		return FreeStackpVector(stkh, i);
	    
	return stkh;
	
} /* end of NewStackpVector */


/* -------------------------------------------- */
/* liberation d'un vecteur de stacks		*/
/*						*/
/* @function: FreeStackpVector			*/
/* -------------------------------------------- */

StackpHdle FreeStackpVector(stkh, vectSize)
	StackpHdle   stkh;
	Int32	     vectSize;
{
	Int32	i;
	
	if (stkh) {
	    for (i = 0 ; i < vectSize ; i++)
	        (void) FreeStackp(stkh[i]);
	    FREE(stkh);
	}

	return NULL;
		
} /* end of FreeStackpVector */


/* -------------------------------------------- */
/* resize d'un stack			 	*/
/*						*/
/* @function: ResizeStackp			*/
/* -------------------------------------------- */

Int32 ResizeStackp(stkh, size)
	StackpHdle stkh;
	Int32 size;
{
	Int32 resize = 0;		/* assume error		*/
    	Ptr   *val;
    		 
	if (val = REALLOC(Ptr, (*stkh)->val, size)) {
	    (*stkh)->size = resize = size;
	    (*stkh)->val = val;
	}

	if (! resize)
	    sStkpLastError = kStkpMemErr;

	return resize;
	
} /* end of ResizeStackp */

/* -------------------------------------------- */
/* empilage(/lement) 				*/
/*						*/
/* @function: PushpIn				*/
/* -------------------------------------------- */

Bool PushpIn(stkh, val)
	StackpHdle stkh;
	Ptr val;	
{
	if (((*stkh)->top >= (*stkh)->size) && (! ExpandStack(stkh)))
	    return Faux;

	(*stkh)->val[((*stkh)->top)++] = val;

	return Vrai;
	
} /* end of PushpIn */

/* -------------------------------------------- */
/* depilage(/lement)				*/
/*						*/
/* @function: PoppOut				*/
/* -------------------------------------------- */

Bool PoppOut(stkh, val)
	StackpHdle stkh;
	Ptr *val;
{
	if ((*stkh)->top <= 0)
	    return Faux;
	
	*val = (*stkh)->val[--((*stkh)->top)];

	if (    ((*stkh)->top < ((*stkh)->size >> 1)) 
	     && ((*stkh)->top > kMinStackpSize))

	    (void) ShrinkStack(stkh);

	return Vrai;
	
} /* end of PoppOut */
		
/* -------------------------------------------- */
/* lecture descendante				*/
/*						*/
/* @function: ReadpDown				*/
/* -------------------------------------------- */

Bool ReadpDown(stkp, val)
	StackpPtr stkp;
	Ptr 	  *val;
{
	if (stkp->cursor <= 0)
	    return Faux;
		
	*val = stkp->val[--(stkp->cursor)];
	
	return Vrai;
	
} /* end of ReadpDown */

/* -------------------------------------------- */
/* lecture ascendante				*/
/*						*/
/* @function: ReadpUp				*/
/* -------------------------------------------- */

Bool ReadpUp(stkp, val)
	StackpPtr stkp;
	Ptr       *val;	
{
	if (stkp->cursor >= stkp->top)
	    return Faux;
		
	*val = stkp->val[(stkp->cursor)++];
	
	return Vrai;
	
} /* end of ReadpUp */

/* -------------------------------------------- */
/* remontee/descente du curseur			*/
/*						*/
/* @function: CurspToTop			*/
/* @function: CurspToBottom			*/
/* -------------------------------------------- */

void CurspToTop(stkp)
	StackpPtr stkp;
{
	stkp->cursor = stkp->top;

} /* end of CurspToTop */

void CurspToBottom(stkp)
	StackpPtr stkp;
{
	stkp->cursor = 0;

} /* end of CurspToBottom */

/* -------------------------------------------- */
/* echange des valeurs cursor <-> (top - 1) 	*/
/*						*/
/* @function: CurspSwap				*/
/* -------------------------------------------- */

void CurspSwap(stkp)
	StackpPtr stkp;
{
	Ptr	tmp;
	
	if ((stkp->top <= 0) || (stkp->cursor < 0))
	    return;
	
	tmp = stkp->val[stkp->cursor];
	stkp->val[stkp->cursor] = stkp->val[stkp->top - 1];
	stkp->val[stkp->top - 1] = tmp;
	
} /* end of CurspSwap */

/* -------------------------------------------- */
/* Recherche d'une valeur en stack a partir du	*/
/* curseur courant en descendant.		*/
/* on laisse le curseur a l'endroit trouve	*/
/*						*/
/* @function: SearchDownStackp			*/
/* -------------------------------------------- */

Bool SearchDownStackp(stkp, sval)
	StackpPtr stkp;
	Ptr       sval;
{
	Ptr	val;
	Bool	more;
	
	while (more = ReadpDown(stkp, &val))
	    if (val == sval) 
		break;
	
	return more;
	
} /* end of SearchDownStackp */

/* -------------------------------------------- */
/* inverse l'ordre des elements dans un stack	*/
/*						*/
/* @function: ReverseStackp			*/
/* -------------------------------------------- */

Bool ReverseStackp(stkp)
	StackpPtr stkp;
{
	Ptr	*t, *b, swp;

	if (stkp->top <= 0) 
	    return Faux;
	
	b = stkp->val;
	t = b + stkp->top - 1;

	while (t > b) {
	     swp  = *t;
	     *t-- = *b;
	     *b++ = swp;
	}

	return Vrai;
	
} /* end of ReverseStackp */   
